@@ -18,9 +18,9 @@ import contextlib
 import itertools
 import sys
 
-import fixtures
+import mock
 from mox3 import mox
-from oslo.utils import encodeutils
+from oslo_utils import encodeutils
 from oslotest import base
 import requests
 import six
@@ -37,6 +37,7 @@ API_VERSION = "2.0"
 FORMAT = 'json'
 TOKEN = 'testtoken'
 ENDURL = 'localurl'
+REQUEST_ID = 'test_request_id'
 
 
 @contextlib.contextmanager
@@ -50,7 +51,7 @@ def capture_std_streams():
         sys.stdout, sys.stderr = stdout, stderr
 
 
-class FakeStdout:
+class FakeStdout(object):
 
     def __init__(self):
         self.content = []
@@ -61,11 +62,11 @@ class FakeStdout:
     def make_string(self):
         result = ''
         for line in self.content:
-            result = result + line
+            result += encodeutils.safe_decode(line, 'utf-8')
         return result
 
 
-class MyResp(object):
+class MyResp(requests.Response):
     def __init__(self, status_code, headers=None, reason=None):
         self.status_code = status_code
         self.headers = headers or {}
@@ -91,10 +92,14 @@ class MyUrlComparator(mox.Comparator):
         lhsp = urlparse.urlparse(self.lhs)
         rhsp = urlparse.urlparse(rhs)
 
+        lhs_qs = urlparse.parse_qsl(lhsp.query)
+        rhs_qs = urlparse.parse_qsl(rhsp.query)
+
         return (lhsp.scheme == rhsp.scheme and
                 lhsp.netloc == rhsp.netloc and
                 lhsp.path == rhsp.path and
-                urlparse.parse_qs(lhsp.query) == urlparse.parse_qs(rhsp.query))
+                len(lhs_qs) == len(rhs_qs) and
+                set(lhs_qs) == set(rhs_qs))
 
     def __str__(self):
         if self.client and self.client.format != FORMAT:
@@ -171,57 +176,61 @@ class CLITestV20Base(base.BaseTestCase):
     test_id = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'
     id_field = 'id'
 
+    non_admin_status_resources = []
+
     def _find_resourceid(self, client, resource, name_or_id,
                          cmd_resource=None, parent_id=None):
         return name_or_id
 
-    def _get_attr_metadata(self):
-        return self.metadata
-
     def setUp(self, plurals=None):
         """Prepare the test environment."""
         super(CLITestV20Base, self).setUp()
         client.Client.EXTED_PLURALS.update(constants.PLURALS)
         if plurals is not None:
             client.Client.EXTED_PLURALS.update(plurals)
-        self.metadata = {'plurals': client.Client.EXTED_PLURALS,
-                         'xmlns': constants.XML_NS_V20,
-                         constants.EXT_NS: {'prefix':
-                                            'http://xxxx.yy.com'}}
+        self.metadata = {'plurals': client.Client.EXTED_PLURALS}
         self.mox = mox.Mox()
         self.endurl = ENDURL
         self.fake_stdout = FakeStdout()
-        self.useFixture(fixtures.MonkeyPatch('sys.stdout', self.fake_stdout))
-        self.useFixture(fixtures.MonkeyPatch(
-            'neutronclient.neutron.v2_0.find_resourceid_by_name_or_id',
-            self._find_resourceid))
-        self.useFixture(fixtures.MonkeyPatch(
-            'neutronclient.neutron.v2_0.find_resourceid_by_id',
-            self._find_resourceid))
-        self.useFixture(fixtures.MonkeyPatch(
-            'neutronclient.v2_0.client.Client.get_attr_metadata',
-            self._get_attr_metadata))
+
+        self.addCleanup(mock.patch.stopall)
+        mock.patch('sys.stdout', new=self.fake_stdout).start()
+        mock.patch('neutronclient.neutron.v2_0.find_resourceid_by_name_or_id',
+                   new=self._find_resourceid).start()
+        mock.patch('neutronclient.neutron.v2_0.find_resourceid_by_id',
+                   new=self._find_resourceid).start()
+
         self.client = client.Client(token=TOKEN, endpoint_url=self.endurl)
+        self.client.format = self.format
+
+    def register_non_admin_status_resource(self, resource_name):
+        # TODO(amotoki):
+        # It is recommended to define
+        # "non_admin_status_resources in each test class rather than
+        # using register_non_admin_status_resource method.
+
+        # If we change self.non_admin_status_resources like this,
+        # we need to ensure this should be an instance variable
+        # to avoid changing the class variable.
+        if (id(self.non_admin_status_resources) ==
+                id(self.__class__.non_admin_status_resources)):
+            self.non_admin_status_resources = (self.__class__.
+                                               non_admin_status_resources[:])
+        self.non_admin_status_resources.append(resource_name)
 
     def _test_create_resource(self, resource, cmd, name, myid, args,
                               position_names, position_values,
                               tenant_id=None, tags=None, admin_state_up=True,
                               extra_body=None, cmd_resource=None,
-                              parent_id=None, **kwargs):
+                              parent_id=None, no_api_call=False,
+                              expected_exception=None,
+                              **kwargs):
         self.mox.StubOutWithMock(cmd, "get_client")
         self.mox.StubOutWithMock(self.client.httpclient, "request")
         cmd.get_client().MultipleTimes().AndReturn(self.client)
-        non_admin_status_resources = ['subnet', 'floatingip', 'security_group',
-                                      'security_group_rule', 'qos_queue',
-                                      'network_gateway', 'gateway_device',
-                                      'credential', 'network_profile',
-                                      'policy_profile', 'ikepolicy',
-                                      'ipsecpolicy', 'metering_label',
-                                      'metering_label_rule', 'net_partition',
-                                      'fox_socket']
         if not cmd_resource:
             cmd_resource = resource
-        if (resource in non_admin_status_resources):
+        if (resource in self.non_admin_status_resources):
             body = {resource: {}, }
         else:
             body = {resource: {'admin_state_up': admin_state_up, }, }
@@ -239,7 +248,6 @@ class CLITestV20Base(base.BaseTestCase):
                 {self.id_field: myid}, }
         if name:
             ress[resource].update({'name': name})
-        self.client.format = self.format
         resstr = self.client.serialize(ress)
         # url method body
         resource_plural = neutronV2_0._get_resource_plural(cmd_resource,
@@ -247,27 +255,27 @@ class CLITestV20Base(base.BaseTestCase):
         path = getattr(self.client, resource_plural + "_path")
         if parent_id:
             path = path % parent_id
-        # Work around for LP #1217791. XML deserializer called from
-        # MyComparator does not decodes XML string correctly.
-        if self.format == 'json':
-            mox_body = MyComparator(body, self.client)
-        else:
-            mox_body = self.client.serialize(body)
-        self.client.httpclient.request(
-            end_url(path, format=self.format), 'POST',
-            body=mox_body,
-            headers=mox.ContainsKeyValue(
-                'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr))
-        args.extend(['--request-format', self.format])
+        mox_body = MyComparator(body, self.client)
+
+        if not no_api_call:
+            self.client.httpclient.request(
+                end_url(path, format=self.format), 'POST',
+                body=mox_body,
+                headers=mox.ContainsKeyValue(
+                    'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr))
         self.mox.ReplayAll()
         cmd_parser = cmd.get_parser('create_' + resource)
-        shell.run_command(cmd, cmd_parser, args)
+        if expected_exception:
+            self.assertRaises(expected_exception,
+                              shell.run_command, cmd, cmd_parser, args)
+        else:
+            shell.run_command(cmd, cmd_parser, args)
+            _str = self.fake_stdout.make_string()
+            self.assertIn(myid, _str)
+            if name:
+                self.assertIn(name, _str)
         self.mox.VerifyAll()
         self.mox.UnsetStubs()
-        _str = self.fake_stdout.make_string()
-        self.assertIn(myid, _str)
-        if name:
-            self.assertIn(name, _str)
 
     def _test_list_columns(self, cmd, resources,
                            resources_out, args=('-f', 'json'),
@@ -275,7 +283,6 @@ class CLITestV20Base(base.BaseTestCase):
         self.mox.StubOutWithMock(cmd, "get_client")
         self.mox.StubOutWithMock(self.client.httpclient, "request")
         cmd.get_client().MultipleTimes().AndReturn(self.client)
-        self.client.format = self.format
         if not cmd_resources:
             cmd_resources = resources
 
@@ -289,7 +296,6 @@ class CLITestV20Base(base.BaseTestCase):
             body=None,
             headers=mox.ContainsKeyValue(
                 'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr))
-        args = tuple(args) + ('--request-format', self.format)
         self.mox.ReplayAll()
         cmd_parser = cmd.get_parser("list_" + cmd_resources)
         shell.run_command(cmd, cmd_parser, args)
@@ -300,7 +306,7 @@ class CLITestV20Base(base.BaseTestCase):
                              fields_1=(), fields_2=(), page_size=None,
                              sort_key=(), sort_dir=(), response_contents=None,
                              base_args=None, path=None, cmd_resources=None,
-                             parent_id=None, output_format=None):
+                             parent_id=None, output_format=None, query=""):
         self.mox.StubOutWithMock(cmd, "get_client")
         self.mox.StubOutWithMock(self.client.httpclient, "request")
         cmd.get_client().MultipleTimes().AndReturn(self.client)
@@ -312,14 +318,11 @@ class CLITestV20Base(base.BaseTestCase):
         else:
             contents = response_contents
         reses = {resources: contents}
-        self.client.format = self.format
         resstr = self.client.serialize(reses)
         # url method body
-        query = ""
         args = base_args if base_args is not None else []
         if detail:
             args.append('-D')
-        args.extend(['--request-format', self.format])
         if fields_1:
             for field in fields_1:
                 args.append('--fields')
@@ -402,7 +405,7 @@ class CLITestV20Base(base.BaseTestCase):
     def _test_list_resources_with_pagination(self, resources, cmd,
                                              base_args=None,
                                              cmd_resources=None,
-                                             parent_id=None):
+                                             parent_id=None, query=""):
         self.mox.StubOutWithMock(cmd, "get_client")
         self.mox.StubOutWithMock(self.client.httpclient, "request")
         cmd.get_client().MultipleTimes().AndReturn(self.client)
@@ -419,11 +422,10 @@ class CLITestV20Base(base.BaseTestCase):
                                             'rel': 'next'}]}
         reses2 = {resources: [{'id': 'myid3', },
                               {'id': 'myid4', }]}
-        self.client.format = self.format
         resstr1 = self.client.serialize(reses1)
         resstr2 = self.client.serialize(reses2)
         self.client.httpclient.request(
-            end_url(path, "", format=self.format), 'GET',
+            end_url(path, query, format=self.format), 'GET',
             body=None,
             headers=mox.ContainsKeyValue(
                 'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr1))
@@ -436,7 +438,6 @@ class CLITestV20Base(base.BaseTestCase):
         self.mox.ReplayAll()
         cmd_parser = cmd.get_parser("list_" + cmd_resources)
         args = base_args if base_args is not None else []
-        args.extend(['--request-format', self.format])
         shell.run_command(cmd, cmd_parser, args)
         self.mox.VerifyAll()
         self.mox.UnsetStubs()
@@ -455,13 +456,8 @@ class CLITestV20Base(base.BaseTestCase):
             path = path % (parent_id, myid)
         else:
             path = path % myid
-        self.client.format = self.format
-        # Work around for LP #1217791. XML deserializer called from
-        # MyComparator does not decodes XML string correctly.
-        if self.format == 'json':
-            mox_body = MyComparator(body, self.client)
-        else:
-            mox_body = self.client.serialize(body)
+        mox_body = MyComparator(body, self.client)
+
         self.client.httpclient.request(
             MyUrlComparator(end_url(path, format=self.format),
                             self.client),
@@ -469,7 +465,6 @@ class CLITestV20Base(base.BaseTestCase):
             body=mox_body,
             headers=mox.ContainsKeyValue(
                 'X-Auth-Token', TOKEN)).AndReturn((MyResp(204), None))
-        args.extend(['--request-format', self.format])
         self.mox.ReplayAll()
         cmd_parser = cmd.get_parser("update_" + cmd_resource)
         shell.run_command(cmd, cmd_parser, args)
@@ -490,7 +485,6 @@ class CLITestV20Base(base.BaseTestCase):
         expected_res = {resource:
                         {self.id_field: myid,
                          'name': 'myname', }, }
-        self.client.format = self.format
         resstr = self.client.serialize(expected_res)
         path = getattr(self.client, cmd_resource + "_path")
         if parent_id:
@@ -502,7 +496,6 @@ class CLITestV20Base(base.BaseTestCase):
             body=None,
             headers=mox.ContainsKeyValue(
                 'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr))
-        args.extend(['--request-format', self.format])
         self.mox.ReplayAll()
         cmd_parser = cmd.get_parser("show_" + cmd_resource)
         shell.run_command(cmd, cmd_parser, args)
@@ -529,7 +522,6 @@ class CLITestV20Base(base.BaseTestCase):
             body=None,
             headers=mox.ContainsKeyValue(
                 'X-Auth-Token', TOKEN)).AndReturn((MyResp(204), None))
-        args.extend(['--request-format', self.format])
         self.mox.ReplayAll()
         cmd_parser = cmd.get_parser("delete_" + cmd_resource)
         shell.run_command(cmd, cmd_parser, args)
@@ -552,7 +544,6 @@ class CLITestV20Base(base.BaseTestCase):
             body=MyComparator(body, self.client),
             headers=mox.ContainsKeyValue(
                 'X-Auth-Token', TOKEN)).AndReturn((MyResp(204), retval))
-        args.extend(['--request-format', self.format])
         self.mox.ReplayAll()
         cmd_parser = cmd.get_parser("delete_" + cmd_resource)
         shell.run_command(cmd, cmd_parser, args)
@@ -562,9 +553,83 @@ class CLITestV20Base(base.BaseTestCase):
         self.assertIn(myid, _str)
 
 
+class TestListCommand(neutronV2_0.ListCommand):
+    resource = 'test_resource'
+    filter_attrs = [
+        'name',
+        'admin_state_up',
+        {'name': 'foo', 'help': 'non-boolean attribute foo'},
+        {'name': 'bar', 'help': 'boolean attribute bar',
+         'boolean': True},
+        {'name': 'baz', 'help': 'integer attribute baz',
+         'argparse_kwargs': {'choices': ['baz1', 'baz2']}},
+    ]
+
+
+class ListCommandTestCase(CLITestV20Base):
+
+    def setUp(self):
+        super(ListCommandTestCase, self).setUp()
+        self.client.extend_list('test_resources', '/test_resources', None)
+        setattr(self.client, 'test_resources_path', '/test_resources')
+
+    def _test_list_resources_filter_params(self, base_args='', query=''):
+        resources = 'test_resources'
+        cmd = TestListCommand(MyApp(sys.stdout), None)
+        self._test_list_resources(resources, cmd,
+                                  base_args=base_args.split(),
+                                  query=query)
+
+    def _test_list_resources_with_arg_error(self, base_args=''):
+        self.addCleanup(self.mox.UnsetStubs)
+        resources = 'test_resources'
+        cmd = TestListCommand(MyApp(sys.stdout), None)
+        # argparse parse error leads to SystemExit
+        self.assertRaises(SystemExit,
+                          self._test_list_resources,
+                          resources, cmd,
+                          base_args=base_args.split())
+
+    def test_list_resources_without_filter(self):
+        self._test_list_resources_filter_params()
+
+    def test_list_resources_use_default_filter(self):
+        self._test_list_resources_filter_params(
+            base_args='--name val1 --admin-state-up False',
+            query='name=val1&admin_state_up=False')
+
+    def test_list_resources_use_custom_filter(self):
+        self._test_list_resources_filter_params(
+            base_args='--foo FOO --bar True',
+            query='foo=FOO&bar=True')
+
+    def test_list_resources_boolean_check_default_filter(self):
+        self._test_list_resources_filter_params(
+            base_args='--admin-state-up True', query='admin_state_up=True')
+        self._test_list_resources_filter_params(
+            base_args='--admin-state-up False', query='admin_state_up=False')
+        self._test_list_resources_with_arg_error(
+            base_args='--admin-state-up non-true-false')
+
+    def test_list_resources_boolean_check_custom_filter(self):
+        self._test_list_resources_filter_params(
+            base_args='--bar True', query='bar=True')
+        self._test_list_resources_filter_params(
+            base_args='--bar False', query='bar=False')
+        self._test_list_resources_with_arg_error(
+            base_args='--bar non-true-false')
+
+    def test_list_resources_argparse_kwargs(self):
+        self._test_list_resources_filter_params(
+            base_args='--baz baz1', query='baz=baz1')
+        self._test_list_resources_filter_params(
+            base_args='--baz baz2', query='baz=baz2')
+        self._test_list_resources_with_arg_error(
+            base_args='--bar non-choice')
+
+
 class ClientV2TestJson(CLITestV20Base):
     def test_do_request_unicode(self):
-        self.client.format = self.format
         self.mox.StubOutWithMock(self.client.httpclient, "request")
         unicode_text = u'\u7f51\u7edc'
         # url with unicode
@@ -579,65 +644,180 @@ class ClientV2TestJson(CLITestV20Base):
         self.client.httpclient.auth_token = encodeutils.safe_encode(
             unicode_text)
         expected_auth_token = encodeutils.safe_encode(unicode_text)
+        resp_headers = {'x-openstack-request-id': REQUEST_ID}
 
         self.client.httpclient.request(
             end_url(expected_action, query=expect_query, format=self.format),
             'PUT', body=expect_body,
             headers=mox.ContainsKeyValue(
                 'X-Auth-Token',
-                expected_auth_token)).AndReturn((MyResp(200), expect_body))
+                expected_auth_token)).AndReturn((MyResp(200, resp_headers),
+                                                 expect_body))
 
         self.mox.ReplayAll()
-        res_body = self.client.do_request('PUT', action, body=body,
-                                          params=params)
+        result = self.client.do_request('PUT', action, body=body,
+                                        params=params)
         self.mox.VerifyAll()
         self.mox.UnsetStubs()
 
         # test response with unicode
-        self.assertEqual(res_body, body)
+        self.assertEqual(body, result)
 
     def test_do_request_error_without_response_body(self):
-        self.client.format = self.format
         self.mox.StubOutWithMock(self.client.httpclient, "request")
         params = {'test': 'value'}
         expect_query = six.moves.urllib.parse.urlencode(params)
         self.client.httpclient.auth_token = 'token'
+        resp_headers = {'x-openstack-request-id': REQUEST_ID}
 
         self.client.httpclient.request(
             MyUrlComparator(end_url(
                 '/test', query=expect_query, format=self.format), self.client),
             'PUT', body='',
             headers=mox.ContainsKeyValue('X-Auth-Token', 'token')
-        ).AndReturn((MyResp(400, reason='An error'), ''))
+        ).AndReturn((MyResp(400, headers=resp_headers, reason='An error'), ''))
 
         self.mox.ReplayAll()
         error = self.assertRaises(exceptions.NeutronClientException,
                                   self.client.do_request, 'PUT', '/test',
                                   body='', params=params)
-        self.assertEqual("An error", str(error))
+        expected_error = "An error\nNeutron server returns " \
+                         "request_ids: %s" % [REQUEST_ID]
+        self.assertEqual(expected_error, str(error))
         self.mox.VerifyAll()
         self.mox.UnsetStubs()
 
+    def test_do_request_with_long_uri_exception(self):
+        long_string = 'x' * 8200                  # 8200 > MAX_URI_LEN:8192
+        params = {'id': long_string}
+        exception = self.assertRaises(exceptions.RequestURITooLong,
+                                      self.client.do_request,
+                                      'GET', '/test', body='', params=params)
+        self.assertNotEqual(0, exception.excess)
 
-class ClientV2UnicodeTestXML(ClientV2TestJson):
-    format = 'xml'
+    def test_do_request_request_ids(self):
+        self.mox.StubOutWithMock(self.client.httpclient, "request")
+        params = {'test': 'value'}
+        expect_query = six.moves.urllib.parse.urlencode(params)
+        self.client.httpclient.auth_token = 'token'
+        body = params
+        expect_body = self.client.serialize(body)
+        resp_headers = {'x-openstack-request-id': REQUEST_ID}
+        self.client.httpclient.request(
+            MyUrlComparator(end_url(
+                '/test', query=expect_query,
+                format=self.format), self.client),
+            'PUT', body=expect_body,
+            headers=mox.ContainsKeyValue('X-Auth-Token', 'token')
+        ).AndReturn((MyResp(200, resp_headers), expect_body))
+
+        self.mox.ReplayAll()
+        result = self.client.do_request('PUT', '/test', body=body,
+                                        params=params)
+        self.mox.VerifyAll()
+        self.mox.UnsetStubs()
+
+        self.assertEqual(body, result)
+        self.assertEqual([REQUEST_ID], result.request_ids)
+
+    def test_list_request_ids_with_retrieve_all_true(self):
+        self.mox.StubOutWithMock(self.client.httpclient, "request")
+
+        path = '/test'
+        resources = 'tests'
+        fake_query = "marker=myid2&limit=2"
+        reses1 = {resources: [{'id': 'myid1', },
+                              {'id': 'myid2', }],
+                  '%s_links' % resources: [{'href': end_url(path, fake_query),
+                                            'rel': 'next'}]}
+        reses2 = {resources: [{'id': 'myid3', },
+                              {'id': 'myid4', }]}
+        resstr1 = self.client.serialize(reses1)
+        resstr2 = self.client.serialize(reses2)
+        resp_headers = {'x-openstack-request-id': REQUEST_ID}
+        self.client.httpclient.request(
+            end_url(path, "", format=self.format), 'GET',
+            body=None,
+            headers=mox.ContainsKeyValue(
+                'X-Auth-Token', TOKEN)).AndReturn((MyResp(200, resp_headers),
+                                                   resstr1))
+        self.client.httpclient.request(
+            MyUrlComparator(end_url(path, fake_query, format=self.format),
+                            self.client), 'GET',
+            body=None,
+            headers=mox.ContainsKeyValue(
+                'X-Auth-Token', TOKEN)).AndReturn((MyResp(200, resp_headers),
+                                                   resstr2))
+        self.mox.ReplayAll()
+        result = self.client.list(resources, path)
+
+        self.mox.VerifyAll()
+        self.mox.UnsetStubs()
+
+        self.assertEqual([REQUEST_ID, REQUEST_ID], result.request_ids)
+
+    def test_list_request_ids_with_retrieve_all_false(self):
+        self.mox.StubOutWithMock(self.client.httpclient, "request")
+
+        path = '/test'
+        resources = 'tests'
+        fake_query = "marker=myid2&limit=2"
+        reses1 = {resources: [{'id': 'myid1', },
+                              {'id': 'myid2', }],
+                  '%s_links' % resources: [{'href': end_url(path, fake_query),
+                                            'rel': 'next'}]}
+        reses2 = {resources: [{'id': 'myid3', },
+                              {'id': 'myid4', }]}
+        resstr1 = self.client.serialize(reses1)
+        resstr2 = self.client.serialize(reses2)
+        resp_headers = {'x-openstack-request-id': REQUEST_ID}
+        self.client.httpclient.request(
+            end_url(path, "", format=self.format), 'GET',
+            body=None,
+            headers=mox.ContainsKeyValue(
+                'X-Auth-Token', TOKEN)).AndReturn((MyResp(200, resp_headers),
+                                                   resstr1))
+        self.client.httpclient.request(
+            MyUrlComparator(end_url(path, fake_query, format=self.format),
+                            self.client), 'GET',
+            body=None,
+            headers=mox.ContainsKeyValue(
+                'X-Auth-Token', TOKEN)).AndReturn((MyResp(200, resp_headers),
+                                                   resstr2))
+        self.mox.ReplayAll()
+        result = self.client.list(resources, path, retrieve_all=False)
+        next(result)
+        self.assertEqual([REQUEST_ID], result.request_ids)
+        next(result)
+        self.assertEqual([REQUEST_ID, REQUEST_ID], result.request_ids)
+        self.mox.VerifyAll()
+        self.mox.UnsetStubs()
 
 
 class CLITestV20ExceptionHandler(CLITestV20Base):
 
     def _test_exception_handler_v20(
-        self, expected_exception, status_code, expected_msg,
-        error_type=None, error_msg=None, error_detail=None,
-        error_content=None):
+            self, expected_exception, status_code, expected_msg,
+            error_type=None, error_msg=None, error_detail=None,
+            request_id=None, error_content=None):
+
+        resp = MyResp(status_code, {'x-openstack-request-id': request_id})
+        if request_id is not None:
+            expected_msg += "\nNeutron server returns " \
+                            "request_ids: %s" % [request_id]
         if error_content is None:
             error_content = {'NeutronError': {'type': error_type,
                                               'message': error_msg,
                                               'detail': error_detail}}
+        expected_content = self.client._convert_into_with_meta(error_content,
+                                                               resp)
 
         e = self.assertRaises(expected_exception,
                               client.exception_handler_v20,
-                              status_code, error_content)
+                              status_code, expected_content)
         self.assertEqual(status_code, e.status_code)
+        self.assertEqual(expected_exception.__name__,
+                         e.__class__.__name__)
 
         if expected_msg is None:
             if error_detail:
@@ -651,7 +831,7 @@ class CLITestV20ExceptionHandler(CLITestV20Base):
                    'fake-network-uuid. The IP address fake-ip is in use.')
         self._test_exception_handler_v20(
             exceptions.IpAddressInUseClient, 409, err_msg,
-            'IpAddressInUse', err_msg, '')
+            'IpAddressInUse', err_msg, '', REQUEST_ID)
 
     def test_exception_handler_v20_neutron_known_error(self):
         known_error_map = [
@@ -668,6 +848,7 @@ class CLITestV20ExceptionHandler(CLITestV20Base):
              exceptions.ExternalIpAddressExhaustedClient, 400),
             ('OverQuota', exceptions.OverQuotaClient, 409),
             ('InvalidIpForNetwork', exceptions.InvalidIpForNetworkClient, 400),
+            ('InvalidIpForSubnet', exceptions.InvalidIpForSubnetClient, 400),
         ]
 
         error_msg = 'dummy exception message'
@@ -676,7 +857,7 @@ class CLITestV20ExceptionHandler(CLITestV20Base):
             self._test_exception_handler_v20(
                 client_exc, status_code,
                 error_msg + '\n' + error_detail,
-                server_exc, error_msg, error_detail)
+                server_exc, error_msg, error_detail, REQUEST_ID)
 
     def test_exception_handler_v20_neutron_known_error_without_detail(self):
         error_msg = 'Network not found'
@@ -684,7 +865,7 @@ class CLITestV20ExceptionHandler(CLITestV20Base):
         self._test_exception_handler_v20(
             exceptions.NetworkNotFoundClient, 404,
             error_msg,
-            'NetworkNotFound', error_msg, error_detail)
+            'NetworkNotFound', error_msg, error_detail, REQUEST_ID)
 
     def test_exception_handler_v20_unknown_error_to_per_code_exception(self):
         for status_code, client_exc in exceptions.HTTP_EXCEPTION_MAP.items():
@@ -693,7 +874,7 @@ class CLITestV20ExceptionHandler(CLITestV20Base):
             self._test_exception_handler_v20(
                 client_exc, status_code,
                 error_msg + '\n' + error_detail,
-                'UnknownError', error_msg, error_detail)
+                'UnknownError', error_msg, error_detail, [REQUEST_ID])
 
     def test_exception_handler_v20_neutron_unknown_status_code(self):
         error_msg = 'Unknown error'
@@ -701,44 +882,52 @@ class CLITestV20ExceptionHandler(CLITestV20Base):
         self._test_exception_handler_v20(
             exceptions.NeutronClientException, 501,
             error_msg + '\n' + error_detail,
-            'UnknownError', error_msg, error_detail)
+            'UnknownError', error_msg, error_detail, REQUEST_ID)
 
     def test_exception_handler_v20_bad_neutron_error(self):
-        error_content = {'NeutronError': {'unknown_key': 'UNKNOWN'}}
-        self._test_exception_handler_v20(
-            exceptions.NeutronClientException, 500,
-            expected_msg={'unknown_key': 'UNKNOWN'},
-            error_content=error_content)
+        for status_code, client_exc in exceptions.HTTP_EXCEPTION_MAP.items():
+            error_content = {'NeutronError': {'unknown_key': 'UNKNOWN'}}
+            self._test_exception_handler_v20(
+                client_exc, status_code,
+                expected_msg="{'unknown_key': 'UNKNOWN'}",
+                error_content=error_content,
+                request_id=REQUEST_ID)
 
     def test_exception_handler_v20_error_dict_contains_message(self):
         error_content = {'message': 'This is an error message'}
-        self._test_exception_handler_v20(
-            exceptions.NeutronClientException, 500,
-            expected_msg='This is an error message',
-            error_content=error_content)
+        for status_code, client_exc in exceptions.HTTP_EXCEPTION_MAP.items():
+            self._test_exception_handler_v20(
+                client_exc, status_code,
+                expected_msg='This is an error message',
+                error_content=error_content,
+                request_id=REQUEST_ID)
 
     def test_exception_handler_v20_error_dict_not_contain_message(self):
+        # 599 is not contained in HTTP_EXCEPTION_MAP.
         error_content = {'error': 'This is an error message'}
-        expected_msg = '%s-%s' % (500, error_content)
+        expected_msg = '%s-%s' % (599, error_content)
         self._test_exception_handler_v20(
-            exceptions.NeutronClientException, 500,
+            exceptions.NeutronClientException, 599,
             expected_msg=expected_msg,
+            request_id=None,
             error_content=error_content)
 
     def test_exception_handler_v20_default_fallback(self):
+        # 599 is not contained in HTTP_EXCEPTION_MAP.
         error_content = 'This is an error message'
-        expected_msg = '%s-%s' % (500, error_content)
+        expected_msg = '%s-%s' % (599, error_content)
         self._test_exception_handler_v20(
-            exceptions.NeutronClientException, 500,
+            exceptions.NeutronClientException, 599,
             expected_msg=expected_msg,
+            request_id=None,
             error_content=error_content)
 
     def test_exception_status(self):
         e = exceptions.BadRequest()
-        self.assertEqual(e.status_code, 400)
+        self.assertEqual(400, e.status_code)
 
         e = exceptions.BadRequest(status_code=499)
-        self.assertEqual(e.status_code, 499)
+        self.assertEqual(499, e.status_code)
 
         # SslCertificateValidationError has no explicit status_code,
         # but should have a 'safe' defined fallback.
@@ -746,7 +935,7 @@ class CLITestV20ExceptionHandler(CLITestV20Base):
         self.assertIsNotNone(e.status_code)
 
         e = exceptions.SslCertificateValidationError(status_code=599)
-        self.assertEqual(e.status_code, 599)
+        self.assertEqual(599, e.status_code)
 
     def test_connection_failed(self):
         self.mox.StubOutWithMock(self.client.httpclient, 'request')
@@ -766,3 +955,60 @@ class CLITestV20ExceptionHandler(CLITestV20Base):
         self.assertIsNotNone(error.status_code)
         self.mox.VerifyAll()
         self.mox.UnsetStubs()
+
+
+class DictWithMetaTest(base.BaseTestCase):
+
+    def test_dict_with_meta(self):
+        body = {'test': 'value'}
+        resp = MyResp(200, {'x-openstack-request-id': REQUEST_ID})
+        obj = client._DictWithMeta(body, resp)
+        self.assertEqual(body, obj)
+
+        # Check request_ids attribute is added to obj
+        self.assertTrue(hasattr(obj, 'request_ids'))
+        self.assertEqual([REQUEST_ID], obj.request_ids)
+
+
+class TupleWithMetaTest(base.BaseTestCase):
+
+    def test_tuple_with_meta(self):
+        body = ('test', 'value')
+        resp = MyResp(200, {'x-openstack-request-id': REQUEST_ID})
+        obj = client._TupleWithMeta(body, resp)
+        self.assertEqual(body, obj)
+
+        # Check request_ids attribute is added to obj
+        self.assertTrue(hasattr(obj, 'request_ids'))
+        self.assertEqual([REQUEST_ID], obj.request_ids)
+
+
+class StrWithMetaTest(base.BaseTestCase):
+
+    def test_str_with_meta(self):
+        body = "test_string"
+        resp = MyResp(200, {'x-openstack-request-id': REQUEST_ID})
+        obj = client._StrWithMeta(body, resp)
+        self.assertEqual(body, obj)
+
+        # Check request_ids attribute is added to obj
+        self.assertTrue(hasattr(obj, 'request_ids'))
+        self.assertEqual([REQUEST_ID], obj.request_ids)
+
+
+class GeneratorWithMetaTest(base.BaseTestCase):
+
+    body = {'test': 'value'}
+    resp = MyResp(200, {'x-openstack-request-id': REQUEST_ID})
+
+    def _pagination(self, collection, path, **params):
+        obj = client._DictWithMeta(self.body, self.resp)
+        yield obj
+
+    def test_generator(self):
+        obj = client._GeneratorWithMeta(self._pagination, 'test_collection',
+                                        'test_path', test_args='test_args')
+        self.assertEqual(self.body, next(obj))
+
+        self.assertTrue(hasattr(obj, 'request_ids'))
+        self.assertEqual([REQUEST_ID], obj.request_ids)
